@@ -28,7 +28,6 @@
 
 #include <glib/gi18n.h>
 #include <gio/gio.h>
-#include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -78,6 +77,18 @@ typedef struct {
 	char *remote_addr_6;
 	char *netmask_6;
 
+	/* sudo on remote command */
+	gboolean sudo;
+
+	/* Socks mode - no tunnel activated */
+	gboolean no_tunnel;
+	gboolean no_remote_command;
+	gboolean network_config_sent;
+	char* no_tunnel_interface;
+	char* socks_bind_address;
+	char* local_bind_address;
+	char* remote_bind_address;
+
 	/* fds for handling input/output of the SSH process */
 	GIOChannel *ssh_stdin_channel;
 	GIOChannel *ssh_stderr_channel;
@@ -117,12 +128,15 @@ static ValidProperty valid_properties[] = {
 	{ NM_SSH_KEY_TAP_DEV,              G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_SSH_KEY_REMOTE_USERNAME,      G_TYPE_STRING, 0, 0, FALSE },
 	/* FIXME should fix host validation for IPv6 addresses */
+	{ NM_SSH_KEY_SUDO,                 G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_SSH_KEY_IP_6,                 G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_SSH_KEY_REMOTE_IP_6,          G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_SSH_KEY_LOCAL_IP_6,           G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_SSH_KEY_NETMASK_6,            G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_SSH_KEY_AUTH_TYPE,            G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_SSH_KEY_KEY_FILE,             G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_SSH_KEY_NO_TUNNEL_INTERFACE,  G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_SSH_KEY_SOCKS_BIND_ADDRESS,   G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_SSH_KEY_PASSWORD"-flags",     G_TYPE_STRING, 0, 0, FALSE },
 	{ NULL,                            G_TYPE_NONE, FALSE }
 };
@@ -162,14 +176,14 @@ validate_one_property (const char *key, const char *value, gpointer user_data)
 	info->have_items = TRUE;
 
 	/* 'name' is the setting name; always allowed but unused */
-	if (!strncmp (key, NM_SETTING_NAME, strlen(NM_SETTING_NAME)))
+	if (!strncmp (key, NM_SETTING_NAME, strlen(NM_SETTING_NAME)) && strlen(key) == strlen(NM_SETTING_NAME))
 		return;
 
 	for (i = 0; info->table[i].name; i++) {
 		ValidProperty prop = info->table[i];
 		long int tmp;
 
-		if (strncmp (prop.name, key, strlen(prop.name)))
+		if (!strncmp (prop.name, key, strlen(prop.name)) && strlen(prop.name) == strlen(key))
 			continue;
 
 		switch (prop.type) {
@@ -310,7 +324,7 @@ resolve_hostname (const char *hostname)
 	const char *p;
 	gboolean is_name = FALSE;
 
-	/* Check if it seems to be a hostname hostname */
+	/* Check if it seems to be a hostname */
 	p = hostname;
 	while (*p) {
 		if (*p != '.' && !isdigit (*p)) {
@@ -377,6 +391,11 @@ send_network_config (NMSshPlugin *plugin)
 	char               *device;
 	char               *resolved_hostname;
 
+	if (io_data->network_config_sent)
+	{
+		return TRUE;
+	}
+
 	g_variant_builder_init (&config, G_VARIANT_TYPE_VARDICT);
 	g_variant_builder_init (&ip4config, G_VARIANT_TYPE_VARDICT);
 	g_variant_builder_init (&ip6config, G_VARIANT_TYPE_VARDICT);
@@ -412,17 +431,26 @@ send_network_config (NMSshPlugin *plugin)
 	else
 		g_warning ("remote_gw unset.");
 
-	/* device */
-	if (io_data->local_dev_number != -1)
+	if (io_data->no_tunnel)
 	{
-		device =
-			(gpointer) g_strdup_printf ("%s%d", io_data->dev_type, io_data->local_dev_number);
-		val = str_to_gvariant (device, FALSE);
-		g_free(device);
+		val = str_to_gvariant (priv->io_data->no_tunnel_interface, FALSE);
 		g_variant_builder_add (&config, "{sv}", NM_VPN_PLUGIN_CONFIG_TUNDEV, val);
 	}
 	else
-		g_warning ("local_dev_number unset.");
+	{
+		/* device */
+		if (io_data->local_dev_number != -1)
+		{
+			device = (gpointer) g_strdup_printf ("%s%d", io_data->dev_type, io_data->local_dev_number);
+			val = str_to_gvariant (device, FALSE);
+			g_free(device);
+			g_variant_builder_add (&config, "{sv}", NM_VPN_PLUGIN_CONFIG_TUNDEV, val);
+		}
+		else
+		{
+			g_warning ("local_dev_number unset.");
+		}
+	}
 
 	/* mtu */
 	if (io_data->mtu > 0)
@@ -509,6 +537,11 @@ send_network_config (NMSshPlugin *plugin)
 
 	/* ---------------------------------------------------- */
 
+	if (io_data->no_tunnel)
+	{
+		g_variant_builder_add (&ip4config, "{sv}", NM_VPN_PLUGIN_IP4_CONFIG_NEVER_DEFAULT, g_variant_new_boolean (TRUE));
+		g_variant_builder_add (&ip6config, "{sv}", NM_VPN_PLUGIN_IP4_CONFIG_NEVER_DEFAULT, g_variant_new_boolean (TRUE));
+	}
 
 	/* Send general config */
 	nm_vpn_service_plugin_set_config ((NMVpnServicePlugin*) plugin,
@@ -523,6 +556,8 @@ send_network_config (NMSshPlugin *plugin)
 	nm_vpn_service_plugin_set_ip4_config ((NMVpnServicePlugin*) plugin,
 	                                      g_variant_builder_end (&ip4config));
 
+	io_data->network_config_sent = TRUE;
+
 	return TRUE;
 }
 
@@ -562,6 +597,12 @@ nm_ssh_stderr_cb (GIOChannel *source, GIOCondition condition, gpointer user_data
 			send_network_config((NMSshPlugin *)plugin);
 		else if(debug)
 			g_message("Not starting local timer because plugin is in STOPPED state");
+	} else if (g_str_has_prefix (str, "debug1: Local forwarding listening on")) {
+		g_message("Port forwarding mode detected");
+		if (priv->io_data->no_remote_command) {
+			g_message("Sending network configuration (no remote_command)");
+			send_network_config((NMSshPlugin *)plugin);
+		}
 	} else if (g_str_has_prefix (str, "debug1: Remote: Server has rejected tunnel device forwarding")) {
 		/* Opening of remote tun device failed... :( */
 		g_warning("Tunnel device open failed on remote server.");
@@ -592,7 +633,7 @@ nm_ssh_get_free_device (const char *device_type)
 
 	for (device = 0; device <= 255; device++)
 	{
-		system_cmd = (gpointer) g_strdup_printf ("%s %s%d >& /dev/null", IFCONFIG, device_type, device);
+		system_cmd = (gpointer) g_strdup_printf ("%s link show %s%d > /dev/null 2>&1", IPROUTE2, device_type, device);
 		if (system(system_cmd) != 0)
 		{
 			g_free(system_cmd);
@@ -774,11 +815,12 @@ nm_ssh_start_ssh_binary (NMSshPlugin *plugin,
 	/* This giant function is basically taking care of passing all the
 	 * correct parameters to ssh (and sshpass) */
 	NMSshPluginPrivate *priv = NM_SSH_PLUGIN_GET_PRIVATE (plugin);
+	char *sudo = NULL;
 	const char *ssh_binary = NULL, *sshpass_binary = NULL, *tmp = NULL;
 	const char *remote = NULL, *port = NULL, *mtu = NULL, *ssh_agent_socket = NULL, *auth_type = NULL;
 	char *known_hosts_file = NULL;
 	char *tmp_arg = NULL;
-	char *ifconfig_cmd_4 = NULL, *ifconfig_cmd_6 = NULL;
+	char *ip_addr_cmd_4 = NULL, *ip_addr_cmd_6 = NULL, *ip_link_cmd = NULL;
 	char *envp[16];
 	long int tmp_int;
 	GPtrArray *args;
@@ -787,6 +829,8 @@ nm_ssh_start_ssh_binary (NMSshPlugin *plugin,
 	gint ssh_stdin_fd, ssh_stderr_fd;
 	int sshpass_pipe[2];
 	const gchar *password = NULL;
+	gchar **bind_addresses = NULL;
+	gchar **bind_address = NULL;
 
 	/* Find ssh */
 	ssh_binary = nm_find_ssh ();
@@ -801,12 +845,30 @@ nm_ssh_start_ssh_binary (NMSshPlugin *plugin,
 
 	/* Allocate io_data structure */
 	priv->io_data = g_malloc0 (sizeof (NMSshPluginIOData));
+	priv->io_data->network_config_sent = FALSE;
 
 	args = g_ptr_array_new ();
 
 	/* Get auth_type from s_vpn */
 	auth_type = nm_setting_vpn_get_data_item (s_vpn, NM_SSH_KEY_AUTH_TYPE);
 
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_SSH_KEY_NO_REMOTE_COMMAND);
+	if (tmp) {
+		priv->io_data->no_remote_command = TRUE;
+	} else {
+		priv->io_data->no_remote_command = FALSE;
+	}
+
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_SSH_KEY_NO_TUNNEL_INTERFACE);
+	if (tmp) {
+		priv->io_data->no_tunnel = TRUE;
+		priv->io_data->no_tunnel_interface = g_strdup (tmp);
+		priv->io_data->no_remote_command = TRUE;
+	} else {
+		priv->io_data->no_tunnel = FALSE;
+		priv->io_data->no_tunnel_interface = NULL;
+	}
+
 	/* Handle different behaviour for different auth types */
 	envp[0] = NULL;
 	if (!strncmp (auth_type, NM_SSH_AUTH_TYPE_PASSWORD, strlen(NM_SSH_AUTH_TYPE_PASSWORD))) {
@@ -890,7 +952,7 @@ nm_ssh_start_ssh_binary (NMSshPlugin *plugin,
 				return FALSE;
 			}
 		} else if (!strncmp (auth_type, NM_SSH_AUTH_TYPE_SSH_AGENT, strlen(NM_SSH_AUTH_TYPE_SSH_AGENT))) {
-			/* Last but not least, the original nm-ssh default behaviour which
+			/* Last but not least, the original nm-ssh default behaviour
 		 	* which is the sanest of all - SSH_AGENT socket */
 			/* FIXME add all the ssh agent logic here */
 			/* Set SSH_AUTH_SOCK from ssh-agent
@@ -950,23 +1012,36 @@ nm_ssh_start_ssh_binary (NMSshPlugin *plugin,
 	add_ssh_arg (args, "TCPKeepAlive=yes");
 
 	/* Device, either tun or tap */
-	add_ssh_arg (args, "-o");
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_SSH_KEY_TAP_DEV);
-	if (tmp && IS_YES(tmp)) {
-		add_ssh_arg (args, "Tunnel=ethernet");
-		g_strlcpy ((gchar *) &priv->io_data->dev_type, "tap", 4);
+	if (priv->io_data->no_tunnel) {
+		g_message ("Will not establish tunnel (will use dummy interface)");
 	} else {
-		add_ssh_arg (args, "Tunnel=point-to-point");
-		g_strlcpy ((gchar *) &priv->io_data->dev_type, "tun", 4);
+		add_ssh_arg (args, "-o");
+		if (tmp && IS_YES(tmp)) {
+			add_ssh_arg (args, "Tunnel=ethernet");
+			g_strlcpy ((gchar *) &priv->io_data->dev_type, "tap", 4);
+		} else {
+			add_ssh_arg (args, "Tunnel=point-to-point");
+			g_strlcpy ((gchar *) &priv->io_data->dev_type, "tun", 4);
+		}
+
+		/* Get a local tun/tap */
+		priv->io_data->local_dev_number = nm_ssh_get_free_device(priv->io_data->dev_type);
+		if (priv->io_data->local_dev_number == -1)
+		{
+			g_warning("Could not assign a free tun/tap device.");
+			nm_vpn_service_plugin_failure (NM_VPN_SERVICE_PLUGIN (plugin), NM_VPN_PLUGIN_FAILURE_CONNECT_FAILED);
+			return FALSE;
+		}
 	}
 
-	/* Get a local tun/tap */
-	priv->io_data->local_dev_number = nm_ssh_get_free_device(priv->io_data->dev_type);
-	if (priv->io_data->local_dev_number == -1)
-	{
-		g_warning("Could not assign a free tun/tap device.");
-		nm_vpn_service_plugin_failure (NM_VPN_SERVICE_PLUGIN (plugin), NM_VPN_PLUGIN_FAILURE_CONNECT_FAILED);
-		return FALSE;
+	/* Device, either tun or tap */
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_SSH_KEY_SUDO);
+	if (tmp && IS_YES(tmp)) {
+		g_message ("Using '%s' as prefix for remote commands", SUDO);
+		sudo = g_strdup_printf ("%s ", SUDO);
+	} else {
+		sudo = g_strdup ("");
 	}
 
 	/* Remote */
@@ -1034,7 +1109,7 @@ nm_ssh_start_ssh_binary (NMSshPlugin *plugin,
 			free_ssh_args (args);
 			return FALSE;
 		}
-		priv->io_data->remote_dev_number = tmp_int;
+		priv->io_data->remote_dev_number = (gint) tmp_int;
 	} else {
 		/* Use tun100/tap100 by default */
 		priv->io_data->remote_dev_number = NM_SSH_DEFAULT_REMOTE_DEV;
@@ -1139,14 +1214,17 @@ nm_ssh_start_ssh_binary (NMSshPlugin *plugin,
 		priv->io_data->netmask_6 = g_strdup("");
 	}
 
+	if (priv->io_data->no_tunnel) {
+		/* Nothing to do here... */
+	} else {
+		/* The -w option, provide a remote and local tun/tap device */
+		tmp_arg = (gpointer) g_strdup_printf ("TunnelDevice=%d:%d", priv->io_data->local_dev_number, priv->io_data->remote_dev_number);
+		add_ssh_arg (args, "-o");
+		add_ssh_arg (args, tmp_arg);
+		g_free(tmp_arg);
+	}
 
-	/* The -w option, provide a remote and local tun/tap device */
-	tmp_arg = (gpointer) g_strdup_printf (
-			"TunnelDevice=%d:%d", priv->io_data->local_dev_number, priv->io_data->remote_dev_number);
-	add_ssh_arg (args, "-o"); add_ssh_arg (args, tmp_arg);
-	g_free(tmp_arg);
-
-	/* Remote username, should usually be root */
+	/* Remote username. Should usually be root */
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_SSH_KEY_REMOTE_USERNAME);
 	if (tmp && strlen (tmp)) {
 		priv->io_data->username = g_strdup (tmp);
@@ -1159,6 +1237,71 @@ nm_ssh_start_ssh_binary (NMSshPlugin *plugin,
 	add_ssh_arg (args, tmp_arg);
 	g_free(tmp_arg);
 
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_SSH_KEY_SOCKS_BIND_ADDRESS);
+	if (tmp && strlen (tmp)) {
+		priv->io_data->socks_bind_address = g_strdup (tmp);
+	} else {
+		priv->io_data->socks_bind_address = NULL;
+	}
+
+	if (priv->io_data->socks_bind_address)
+	{
+		bind_addresses = g_strsplit_set (priv->io_data->socks_bind_address, " ", 0);
+		for (bind_address = bind_addresses; *bind_address; bind_address++) {
+			if (strlen (*bind_address)) {
+				tmp_arg = g_strdup_printf ("DynamicForward=%s", *bind_address);
+				add_ssh_arg (args, "-o");
+				add_ssh_arg (args, tmp_arg);
+				g_free(tmp_arg);
+			}
+		}
+
+		if (bind_addresses)
+			g_strfreev (bind_addresses);
+	}
+
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_SSH_KEY_LOCAL_BIND_ADDRESS);
+	if (tmp && strlen (tmp)) {
+		priv->io_data->local_bind_address = g_strdup (tmp);
+	} else {
+		priv->io_data->local_bind_address = NULL;
+	}
+
+	if (priv->io_data->local_bind_address)
+	{
+		bind_addresses = g_strsplit_set (priv->io_data->local_bind_address, " ", 0);
+		for (bind_address = bind_addresses; *bind_address; bind_address++) {
+			if (strlen (*bind_address)) {
+				add_ssh_arg (args, "-L");
+				add_ssh_arg (args, *bind_address);
+			}
+		}
+
+		if (bind_addresses)
+			g_strfreev (bind_addresses);
+	}
+
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_SSH_KEY_REMOTE_BIND_ADDRESS);
+	if (tmp && strlen (tmp)) {
+		priv->io_data->remote_bind_address = g_strdup (tmp);
+	} else {
+		priv->io_data->remote_bind_address = NULL;
+	}
+
+	if (priv->io_data->remote_bind_address)
+	{
+		bind_addresses = g_strsplit_set (priv->io_data->remote_bind_address, " ", 0);
+		for (bind_address = bind_addresses; *bind_address; bind_address++) {
+			if (strlen (*bind_address)) {
+				add_ssh_arg (args, "-R");
+				add_ssh_arg (args, *bind_address);
+			}
+		}
+
+		if (bind_addresses)
+			g_strfreev (bind_addresses);
+	}
+
 	/* Connect to remote */
 	tmp_arg = g_strdup_printf ("HostName=%s", priv->io_data->remote_gw);
 	add_ssh_arg (args, "-o");
@@ -1170,34 +1313,60 @@ nm_ssh_start_ssh_binary (NMSshPlugin *plugin,
 	add_ssh_arg (args, "NetworkManager-ssh");
 
 	/* Command line to run on remote machine */
-	ifconfig_cmd_4 = (gpointer) g_strdup_printf (
-		"%s %s%d inet %s netmask %s pointopoint %s mtu %d",
-		IFCONFIG,
+	ip_link_cmd = (gpointer) g_strdup_printf (
+		"%s%s link set mtu %d dev %s%d up",
+		sudo,
+		IPROUTE2,
+		priv->io_data->mtu,
 		priv->io_data->dev_type,
-		priv->io_data->remote_dev_number,
+		priv->io_data->remote_dev_number);
+
+	/* Command line to run on remote machine */
+	ip_addr_cmd_4 = (gpointer) g_strdup_printf (
+		"%s%s addr add %s/%s peer %s/%s dev %s%d",
+		sudo,
+		IPROUTE2,
 		priv->io_data->remote_addr,
 		priv->io_data->netmask,
 		priv->io_data->local_addr,
-		priv->io_data->mtu);
+		priv->io_data->netmask,
+		priv->io_data->dev_type,
+		priv->io_data->remote_dev_number);
 
 	/* IPv6 ifconfig command to run on remote machine */
 	if (priv->io_data->ipv6) {
-		ifconfig_cmd_6 = (gpointer) g_strdup_printf (
-			"%s %s%d add %s/%s",
-			IFCONFIG,
-			priv->io_data->dev_type,
-			priv->io_data->remote_dev_number,
+		ip_addr_cmd_6 = (gpointer) g_strdup_printf (
+			"%s%s addr add %s/%s peer %s/%s dev %s%d",
+			sudo,
+			IPROUTE2,
 			priv->io_data->remote_addr_6,
-			priv->io_data->netmask_6);
+			priv->io_data->netmask_6,
+			priv->io_data->local_addr_6,
+			priv->io_data->netmask_6,
+			priv->io_data->dev_type,
+			priv->io_data->remote_dev_number);
+	}
+
+	if (priv->io_data->no_tunnel || priv->io_data->no_remote_command) {
+		g_message ("Using -N - no remote command set, or SOCKS mode set");
+		add_ssh_arg (args, "-N");
 	} else {
-		ifconfig_cmd_6 = g_strdup("");
+		/* Concatenate ip_addr_cmd_4, ip_addr_cmd_6 and ip_link_cmd to one command */
+		if (ip_addr_cmd_6)
+		{
+			tmp_arg = g_strconcat(ip_addr_cmd_4, "; ", ip_addr_cmd_6, "; ", ip_link_cmd, NULL);
+			g_free(ip_addr_cmd_6);
+		}
+		else
+		{
+			tmp_arg = g_strconcat(ip_addr_cmd_4, "; ", ip_link_cmd, NULL);
+		}
+
+		add_ssh_arg (args, tmp_arg);
+		g_free(ip_addr_cmd_4);
+		g_free(ip_link_cmd);
+		g_free(tmp_arg);
 	}
-	/* Concatenate ifconfig_cmd_4 and ifconfig_cmd_6 to one command */
-	tmp_arg = g_strconcat(ifconfig_cmd_4, "; ", ifconfig_cmd_6, NULL);
-	add_ssh_arg (args, tmp_arg);
-	g_free(ifconfig_cmd_4);
-	g_free(ifconfig_cmd_6);
-	g_free(tmp_arg);
 
 	/* Wrap it up */
 	g_ptr_array_add (args, NULL);
@@ -1231,6 +1400,8 @@ nm_ssh_start_ssh_binary (NMSshPlugin *plugin,
 		close(sshpass_pipe[1]);
 	}
 
+	g_free (sudo);
+
 	g_message ("ssh started with pid %d", pid);
 
 	/* Add a watch for the SSH stdin and stderr */
@@ -1264,7 +1435,7 @@ validate_ssh_agent_socket(const char* ssh_agent_socket, GError **error)
 	GFile           *gfile = NULL;
 	GFileInfo       *info = NULL;
 	if (debug)
-		g_message ("Inspecing ssh agent socket at: '%s'\n", ssh_agent_socket);
+		g_message ("Inspecting ssh agent socket at: '%s'\n", ssh_agent_socket);
 
 	if (!g_file_test (ssh_agent_socket, G_FILE_TEST_EXISTS))
 		return FALSE;
@@ -1306,7 +1477,7 @@ real_connect (NMVpnServicePlugin   *plugin,
 	if (!nm_ssh_properties_validate (s_vpn, error))
 		return FALSE;
 
-	/* Finally try to start SSH */
+	/* Finally, try to start SSH */
 	if (!nm_ssh_start_ssh_binary (NM_SSH_PLUGIN (plugin), s_vpn, user_name, error))
 		return FALSE;
 
@@ -1347,7 +1518,7 @@ real_need_secrets (NMVpnServicePlugin *plugin,
 	if (!auth_type)
 		auth_type = NM_SSH_AUTH_TYPE_SSH_AGENT;
 
-	/* Lets see if we need some passwords... */
+	/* Let's see if we need some passwords... */
 	if (!strncmp (auth_type, NM_SSH_AUTH_TYPE_PASSWORD, strlen(NM_SSH_AUTH_TYPE_PASSWORD))) {
 		/* Password auth */
 		need_secrets = TRUE;
@@ -1545,6 +1716,9 @@ main (int argc, char *argv[])
 	if (system ("/sbin/modprobe tun") == -1)
 		exit (EXIT_FAILURE);
 
+	if (system ("/sbin/modprobe dummy numdummies=8") == -1)
+		exit (EXIT_FAILURE);
+
 	plugin = nm_ssh_plugin_new (bus_name);
 	if (!plugin)
 		exit (EXIT_FAILURE);
